"""Tools for mocking OpenAI API responses (for testing purposes)."""

from __future__ import annotations  # noqa FI58


class MockCompletion:
    """Mock openai completion object."""

    def __init__(self, content: str | None = None, responses_per_request: int = 1):
        """Initialize a new instance of `MockCompletion` class.

        Args:
            content: The mocked content to be returned, i.e.,
                `json.dumps({"comment": "This is a great movie!",
                "label": 1})`.
            responses_per_request: Number of responses
                for each request.
        """
        # We generate 5 identical responses for each API call by default.
        if content is not None:
            # Mock an OpenAI ChatCompletion with identical responses.
            self.choices = [{"message": {"content": content}}] * responses_per_request
        else:
            # Mock an OpenAI ChatCompletion with different responses.
            # Only used in mock_batch_openai_response_with_different_completion.
            # The choice will be replaced later in the function.
            self.choices = []

    def __repr__(self):
        """Return a string representation.

        Returns:
            _string: A string representation of the object, including its choices.
        """
        _string = f"<MockObject choices={self.choices}>"
        return _string


class MockBatchDifferentCompletions:
    """Mock batch openai completion object."""

    def __init__(self, length: int = 4) -> None:
        """Init a new instance of `MockBatchDifferentCompletions`.

        Args:
            length: Length of the batch completions.

        This class is designed to simulate the response of ChatGPTAgent and test the
        generation process of the OpenAIDataSetGenerator with
        `filter_duplicated_examples` set to True in
        `dataset_generator_with_filter_test`.

        The class works in conjunction with OpenAIDataSetGenerator with
        batch_size = 2, responses_per_request = 3, expected_num_examples
        = 5, and filter_duplicated_examples = True.

        Explanation of the generation process:

        In the first API call, the generator produces 2 * 3 = 6 responses.
        After filtering duplicates, the generated_dataset will be:

        Dataset.from_dict({
            "input_col": ["1", "2"],
            "output_col": ["a", "a"],
        })

        The second API call reduces batch_size to 1 and generates 3 more
        responses. After filtering duplicates, the generated_dataset will be:

        Dataset.from_dict({
            "input_col": ["1", "2", "3"],
            "output_col": ["a", "a", "a"],
        })

        The third API call again uses batch_size = 1 and generates another
        3 responses. After filtering duplicates, the generated_dataset will be:

        Dataset.from_dict({
            "input_col": ["1", "2", "3"],
            "output_col": ["b", "a", "a"],
        })

        The fourth and API call also uses batch_size = 1 and generates the final
        3 responses. After filtering duplicates, the generated_dataset will be:

        Dataset.from_dict({
            "input_col": ["1", "2", "3", "4", "5"],
            "output_col": ["b", "a", "a", "c", "a"],
        })

        The fivth and API call is specifically designed for
        testing generate dataset_dict.
        """
        assert length == 4 or length == 5
        self.mock_completions: list[list[MockCompletion]] = []
        self.current_index = 0
        mock_completion_1 = MockCompletion()
        mock_completion_1.choices = [
            {"message": {"content": '{"input": "1", "output": "a"}'}},
            {"message": {"content": '{"input": "1", "output": "b"}'}},
            {"message": {"content": '{"input": "1", "output": "a"}'}},
        ]
        mock_completion_2 = MockCompletion()
        mock_completion_2.choices = [
            {"message": {"content": '{"input": "1", "output": "c"}'}},
            {"message": {"content": '{"input": "2", "output": "a"}'}},
            {"message": {"content": '{"input": "2", "output": "b"}'}},
        ]
        self.mock_completions.append(
            [
                mock_completion_1,
                mock_completion_2,
            ]
        )
        mock_completion_3 = MockCompletion()
        mock_completion_3.choices = [
            {"message": {"content": '{"input": "3", "output": "a"}'}},
            {"message": {"content": '{"input": "3", "output": "a"}'}},
            {"message": {"content": '{"input": "3", "output": "b"}'}},
        ]
        self.mock_completions.append([mock_completion_3])

        mock_completion_4 = MockCompletion()
        mock_completion_4.choices = [
            {"message": {"content": '{"input": "1", "output": "b"}'}},
            {"message": {"content": '{"input": "1", "output": "b"}'}},
            {"message": {"content": '{"input": "1", "output": "b"}'}},
        ]
        self.mock_completions.append([mock_completion_4])
        mock_completion_5 = MockCompletion()
        mock_completion_5.choices = [
            {"message": {"content": '{"input": "4", "output": "c"}'}},
            {"message": {"content": '{"input": "4", "output": "c"}'}},
            {"message": {"content": '{"input": "5", "output": "a"}'}},
        ]
        self.mock_completions.append([mock_completion_5])
        if length == 5:
            self.mock_completions.append(
                [
                    mock_completion_1,
                    mock_completion_2,
                ]
            )


def mock_batch_openai_response_identical_completions(
    prompts: list[str],
    content: str,
    temperature: float,
    presence_penalty: float = 0,
    frequency_penalty: float = 0,
    responses_per_request: int = 5,
    requests_per_minute: int = 80,
) -> list[MockCompletion]:
    """Generate a batch of mock completion objects.

        This function creates a batch of `MockCompletion`
        object with a `content` attribute set to an LLM completion string.

    Args:
        prompts: A batch of mocked prompts that won't be used.
        content: The example string to be returned.
        temperature: A mocked temperature.
        presence_penalty: A mocked presence penalty.
        frequency_penalty: A mocked frequency penalty.
        responses_per_request: Number of responses for each request.
        requests_per_minute: Number of requests per minute to allow.

    Returns:
        A mock completion object simulating an OpenAI ChatCompletion API response.
    """
    _ = prompts, temperature, presence_penalty, frequency_penalty, requests_per_minute
    mock_completions = [
        MockCompletion(content=content, responses_per_request=responses_per_request)
        for _ in prompts
    ]
    return mock_completions


class UnknownGpt3Exception(Exception):
    """This is a newly-defined exception for testing purposes."""

    pass



"""An interface for automatic model evaluation."""

from __future__ import annotations  # noqa FI58

import json
from abc import ABC, abstractmethod
from typing import Any

import datasets

from prompt2model.model_executor import ModelOutput
from prompt2model.prompt_parser import PromptSpec


class ModelEvaluator(ABC):
    """An interface for automatic model evaluation."""

    @abstractmethod
    def evaluate_model(
        self,
        dataset: datasets.Dataset,
        gt_column: str,
        predictions: list[ModelOutput],
        metrics: list[datasets.Metric] | None = None,
        prompt_spec: PromptSpec | None = None,
    ) -> dict[str, Any]:
        """Evaluate a model on a test set..

        Args:
            dataset: The dataset to evaluate metrics on.
            gt_column: The dataset column to use as ground truth.
            predictions: Model outputs to evaluate.
            metrics: (Optional) The metrics to use, defaults to using chr_f,
                exact_match, and bert_score.
            prompt_spec: (Optional) A PromptSpec to infer the metrics from.

        Returns:
            A dictionary of metric values to return.
        """

    def write_metrics(self, metrics_dict: dict[str, Any], metrics_path: str) -> None:
        """This function writes metrics to a file.

        Args:
            metrics_dict: A dictionary of metrics to write.
            metrics_path: The file path to write metrics to.

        """
        with open(metrics_path, "w") as f:
            json.dump(metrics_dict, f)



"""A dummy class to generate model outputs (for testing purposes)."""

import datasets

from prompt2model.model_executor import ModelExecutor, ModelOutput


class MockModelExecutor(ModelExecutor):
    """An interface for automatic model evaluation."""

    def make_prediction(
        self,
        test_set: datasets.Dataset,
        input_column: str,
    ) -> list[ModelOutput]:
        """Mock the execution of a model on a test set.

        Args:
            test_set: The dataset to make predictions on.
            input_column: The dataset column to use as input to the model.

        Returns:
            An object containing model outputs.
        """
        predictions = []
        for _ in test_set[input_column]:
            model_output = ModelOutput(prediction="", auxiliary_info={})
            predictions.append(model_output)
        return predictions

    def make_single_prediction(self, model_input: str) -> ModelOutput:
        """Mock evaluation on one example.

        Args:
            model_input: The input string to the model.

        Returns:
            A single model output, useful for exposing a model to a user interface.
        """
        _ = model_input
        model_output = ModelOutput(prediction="", auxiliary_info={})
        return model_output



"""Model executor for generative models, including T5-type and GPT-type."""
from __future__ import annotations  # noqa FI58

from typing import Any

import datasets
import torch

from prompt2model.model_executor import ModelExecutor, ModelOutput
from prompt2model.utils import get_formatted_logger

logger = get_formatted_logger("ModelExecutor")


class GenerationModelExecutor(ModelExecutor):
    """Model executor for T5-type and GPT-type models."""

    def generate(
        self,
        input_ids: list[torch.Tensor],
        attention_mask: list[torch.Tensor],
        hyperparameter_choices: dict[str, Any],
    ) -> list[torch.Tensor]:
        """Generates sequences of token IDs using the model.

        Args:
            input_ids: A list of token ID sequences.
            attention_mask: A list of binary masks indicating attended tokens.
            hyperparameter_choices: A dictionary of hyperparameters for inference.

        Returns:
            A list of model output tensors, one for each element in input_ids.
        """
        generate_strategy = hyperparameter_choices.get("generate_strategy", "greedy")
        assert generate_strategy in [
            "beam",  # beam search.
            "top_k",  # top_k sampling.
            "top_p",  # top_p sampling.
            "greedy",  # greedy search.
            "intersect",  # If both top_k and top_p are set, the model will
            # sample from the intersection of the top-k tokens and the top-p tokens.
        ], "Only support top_k/top_p/intersect sampling and beam/greedy search for inference."  # noqa E501
        if generate_strategy == "greedy":
            output = self.model.generate(
                input_ids=input_ids,
                attention_mask=attention_mask,
                max_length=self.sequence_max_length,
                eos_token_id=self.model.config.eos_token_id,
                early_stopping=True,
                repetition_penalty=hyperparameter_choices.get(
                    "repetition_penalty", 2.0
                ),
            )
        elif generate_strategy == "beam":
            output = self.model.generate(
                input_ids=input_ids,
                attention_mask=attention_mask,
                max_length=self.sequence_max_length,
                eos_token_id=self.model.config.eos_token_id,
                early_stopping=True,
                do_sample=False,
                repetition_penalty=hyperparameter_choices.get(
                    "repetition_penalty", 2.0
                ),
                num_beams=hyperparameter_choices.get("num_beams", 3),
            )
        elif generate_strategy == "top_k":
            output = self.model.generate(
                input_ids=input_ids,
                attention_mask=attention_mask,
                max_length=self.sequence_max_length,
                eos_token_id=self.model.config.eos_token_id,
                early_stopping=True,
                do_sample=True,
                repetition_penalty=hyperparameter_choices.get(
                    "repetition_penalty", 2.0
                ),
                top_k=hyperparameter_choices.get("top_k", 20),
            )
        elif generate_strategy == "top_p":
            output = self.model.generate(
                input_ids=input_ids,
                attention_mask=attention_mask,
                max_length=self.sequence_max_length,
                eos_token_id=self.model.config.eos_token_id,
                early_stopping=True,
                do_sample=True,
                repetition_penalty=hyperparameter_choices.get(
                    "repetition_penalty", 2.0
                ),
                top_p=hyperparameter_choices.get("top_p", 0.95),
            )
        else:
            # For intersect sampling.
            output = self.model.generate(
                input_ids=input_ids,
                attention_mask=attention_mask,
                max_length=self.sequence_max_length,
                eos_token_id=self.model.config.eos_token_id,
                early_stopping=True,
                do_sample=True,
                repetition_penalty=hyperparameter_choices.get(
                    "repetition_penalty", 2.0
                ),
                top_k=hyperparameter_choices.get("top_k", 20),
                top_p=hyperparameter_choices.get("top_p", 0.95),
            )
        return output

    def make_prediction(
        self,
        test_set: datasets.Dataset,
        input_column: str,
        hyperparameter_choices: dict[str, Any] = {},
    ) -> list[ModelOutput]:
        """Make predictions with a T5-type or GPT-type model on a test set.

        Args:
            test_set: The dataset to make predictions on. Note that
                make_single_prediction will warp single_model_input
                into a inference_dataset with only one element.
            input_column: The dataset column to use as input to the model.
            hyperparameter_choices: A dictionary of hyperparameter for generate.

        Returns:
            A list of model outputs, one for each element in the test set.
        """
        expected_num_examples = len(test_set)
        model_outputs = []
        longest_input = max(test_set[input_column], key=len)
        if (
            self.tokenizer_max_length is not None
            and len(self.tokenizer.tokenize(longest_input)) > self.tokenizer_max_length
        ):
            logger.warning(
                (
                    "Truncation happened when tokenizing dataset / input string."
                    " You should consider increasing the tokenizer_max_length."
                    " Otherwise the truncation may lead to unexpected results."
                )
            )

        for start_idx in range(0, expected_num_examples, self.batch_size):
            end_idx = min(start_idx + self.batch_size, expected_num_examples)
            batch = datasets.Dataset.from_dict(test_set[start_idx:end_idx])

            input_texts = batch[input_column]
            encoded_inputs = self.tokenizer.batch_encode_plus(
                input_texts,
                truncation=True,
                max_length=self.tokenizer_max_length,
                padding=True,
                return_tensors="pt",
            )
            device = self.model.device
            input_ids = encoded_inputs["input_ids"].to(device)
            attention_mask = encoded_inputs["attention_mask"].to(device)
            output = self.generate(
                input_ids=input_ids,
                attention_mask=attention_mask,
                hyperparameter_choices=hyperparameter_choices,
            )

            for idx, input_text in enumerate(input_texts):
                logits = output[idx]
                decoded_output = self.tokenizer.decode(logits, skip_special_tokens=True)
                model_output = ModelOutput(
                    prediction=decoded_output,
                    auxiliary_info={
                        "input_text": input_text,
                        "logits": logits,
                    },
                )
                model_outputs.append(model_output)

        return model_outputs

    def make_single_prediction(
        self, model_input: str, hyperparameter_choices: dict[str, Any] = {}
    ) -> ModelOutput:
        """Mock evaluation on one example.

        Args:
            model_input: The input string to the model.
            hyperparameter_choices: A dictionary of hyperparameter for inference.

        Returns:
            A single model output, useful for exposing a model to a user interface.
        """
        expected_num_examples = 1
        inference_dataset = datasets.Dataset.from_dict({"model_input": [model_input]})
        inference_column = "model_input"
        assert len(inference_dataset) == expected_num_examples
        model_output = self.make_prediction(
            inference_dataset,
            inference_column,
            hyperparameter_choices,
        )[0]
        return model_output



"""Import all the model executor classes."""

from prompt2model.model_retriever.base import ModelRetriever
from prompt2model.model_retriever.description_based_retriever import (
    DescriptionModelRetriever,
)
from prompt2model.model_retriever.mock import MockModelRetriever

__all__ = ("ModelRetriever", "DescriptionModelRetriever", "MockModelRetriever")



"""Tools for doing efficient similarity search via the Tevatron/faiss libraries."""

import pickle

import numpy as np
from tevatron.faiss_retriever import BaseFaissIPRetriever


def retrieve_objects(
    query_vector: np.ndarray,
    encoded_datasets_path: str,
    document_names: list[str],
    depth: int,
) -> list[tuple[str, float]]:
    """Return a ranked list of object indices and their scores.

    Args:
        query vector: Vector representation of query.
        encoded_datasets_path: Path to file containing encoded dataset index.
        depth: Number of documents to return.

    Returns:
        Ranked list of object names and their inner product similarity to the query.
    """
    assert query_vector.shape[0] == 1, "Only a single query vector is expected."
    assert len(query_vector.shape) == 2, "Query vector must be 1-D."

    with open(encoded_datasets_path, "rb") as f:
        passage_reps, passage_lookup = pickle.load(f)
    retriever = BaseFaissIPRetriever(passage_reps)
    retriever.add(passage_reps)

    all_scores, all_indices = retriever.search(query_vector, depth)
    assert (
        len(all_scores) == len(all_indices) == 1
    ), "Only one query's ranking should be returned."

    psg_scores = all_scores[0]
    ranked_document_names = [document_names[passage_lookup[x]] for x in all_indices[0]]
    score_tuples = list(zip(ranked_document_names, psg_scores))
    return score_tuples



"""This module provides a dummy trainer for testing purposes."""
from __future__ import annotations  # noqa FI58

from typing import Any

import datasets
from transformers import PreTrainedModel  # noqa
from transformers import PreTrainedTokenizer

from prompt2model.model_trainer import BaseTrainer


class MockTrainer(BaseTrainer):
    """This dummy trainer does not actually train anything."""

    def train_model(
        self,
        hyperparameter_choices: dict[str, Any],
        training_datasets: list[datasets.Dataset],
        validation_datasets: list[datasets.Dataset] | None = None,
    ) -> tuple[PreTrainedModel, PreTrainedTokenizer]:
        """This dummy trainer returns the given model without any training.

        Args:
            training_datasets: A list of training datasets.
            hyperparameter_choices: A dictionary of hyperparameters for training.

        Returns:
            A HuggingFace model and tokenizer.
        """
        _ = training_datasets, hyperparameter_choices, validation_datasets
        return self.model, self.tokenizer



"""Import mock classes used in unit tests."""
from test_helpers.dataset_tools import (
    are_dataset_dicts_identical,
    are_datasets_identical,
)
from test_helpers.mock_openai import (
    MockBatchDifferentCompletions,
    MockCompletion,
    UnknownGpt3Exception,
    mock_batch_openai_response_identical_completions,
)
from test_helpers.mock_retrieval import create_test_search_index
from test_helpers.model_and_tokenizer import (
    create_gpt2_model_and_tokenizer,
    create_t5_model_and_tokenizer,
)

__all__ = (
    "MockCompletion",
    "UnknownGpt3Exception",
    "MockBatchDifferentCompletions",
    "create_gpt2_model_and_tokenizer",
    "create_t5_model_and_tokenizer",
    "create_test_search_index",
    "mock_batch_openai_response_identical_completions",
    "are_dataset_dicts_identical",
    "are_datasets_identical",
    "MockBatchResponseDifferentCompletions",
)



"""Retrieve HuggingFace model's size, description, and downloads."""
import json
import os
import re
import subprocess
from pathlib import Path

from huggingface_hub import HfApi


def main(pretrained_model_name: str, cache_dir: str = None) -> None:
    """Downloads and caches a Hugging Face model's metadata.

    Args:
        pretrained_model_name: HuggingFace pretrained_model_name.
        cache_dir: A directory to cache the metadata.
    """
    if cache_dir is None:
        cache_dir = "model_info"
    cache_path = Path.cwd() / cache_dir
    cache_path.mkdir(parents=True, exist_ok=True)

    if len(pretrained_model_name.split("/")) == 2:
        _, model_name = pretrained_model_name.split("/")
    else:
        model_name = pretrained_model_name

    subprocess.run(
        ["git", "clone", f"https://huggingface.co/{pretrained_model_name}"],
        env=dict(os.environ, GIT_LFS_SKIP_SMUDGE="1"),
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    model_bin_file = Path(f"{model_name}/pytorch_model.bin")

    try:
        if model_bin_file.exists():
            with open(model_bin_file, "r") as file:
                content = file.read()
                size = re.search(r"size (\d+)", content).group(1)  # type: ignore
                print(size)
        else:
            model_bin_file = Path(f"{model_name}/pytorch_model.bin.index.json")
            with open(model_bin_file, "r") as file:
                content = json.loads(file.read())  # type: ignore
                size = content["metadata"]["total_size"]  # type: ignore
                print(size)
    except (
        FileNotFoundError,
        PermissionError,
        IOError,
        json.decoder.JSONDecodeError,
    ) as e:
        raise Exception(f"Failed to read {model_name} in {model_bin_file}: {e}")

    with open(f"{model_name}/README.md", "r", encoding="utf-8") as f:
        readme_content = f.read()
        print(readme_content)

    api = HfApi()
    model_meta = api.model_info(pretrained_model_name)
    downloads = model_meta.downloads
    print(pretrained_model_name, downloads)

    model_info = {
        "pretrained_model_name": pretrained_model_name,
        "description": readme_content,
        "size_bytes": size,
        "downloads": downloads,
    }
    model_info_path = Path(f"{cache_dir}/{model_name}.json")
    model_info_path.touch()
    with open(model_info_path, "w") as file:
        file.write(json.dumps(model_info))

    # The model must exist because it can be found by
    # `model_meta = api.model_info(pretrained_model_name)`
    subprocess.run(["rm", "-rf", model_name])


if __name__ == "__main__":
    main("facebook/roscoe-512-roberta-base")
    main("gpt2")



"""Utilities to construct an LLM "metaprompt" for our instruction parser."""

from __future__ import annotations  # noqa FI58

import json

METAPROMPT_INSTRUCTION = """
As a PromptParser, your objective is to carefully analyze prompts and divide them into two distinct components: an 'Instruction' that provides the primary description of the task, and 'Demonstrations' which are optional examples showcasing the task. Your aim is to generate a JSON dictionary response containing the `Instruction` and `Demonstrations` fields, corresponding to these two components. In case there are no demonstrations provided, the 'Demonstrations' field should be marked as 'N/A'. When including demonstrations, only consider complete examples that consist of both input and output pairs, disregarding any incomplete ones. It is crucial to maintain the precise formatting, word choice, and punctuation exactly as presented in the original prompt. Here are some parsed output you can refer to.
"""  # noqa: E501

METAPROMPT_EXAMPLES = [
    (
        """I am trying to cluster entity strings on Wikipedia according to the Wikipedia article title they refer to. To help me with this, for a given entity name, please provide me with a comprehensive set of alternative names that could refer to the same entity. Entities may be weirdly truncated or ambiguous - e.g. "Wind" may refer to the band "Earth, Wind, and Fire" or to "rescue service". For each entity, I will provide you with a sentence where this entity is used to help you understand what this entity refers to. Generate a comprehensive set of alternate entity names as a JSON-formatted list.

Entity: "fictional character"
Context Sentence: "Jenna Marshall is a fictional character created by Sara Shepard for the `` Pretty Little Liars '' book series , and later developed for the Freeform television series adaptation by I. Marlene King and portrayed by Tammin Sursok ."
Alternate Entity Names: ["fictional characters", "characters", "character"]

Entity: "Catholicism"
Context Sentence: "At home , significantly more electorate residents spoke Italian , Cantonese , Mandarin and Greek at home , and whilst the top three religions (Catholicism , no religion and Anglicanism) differed little from other parts of Perth , Buddhism and Eastern Orthodox adherents outnumbered those of the Uniting Church ."
Alternate Entity Names: ["Catholic Church", "Roman Catholic", "Catholic"]

Entity: "Wind"
Context Sentence: "Illinois musicians with a # 1 Billboard Hot 100 hit include artists from the 1950s : Sam Cooke (d. 1964) ; from the 1960s : The Buckinghams ; from the 1970s : Earth , Wind & Fire , The Chi-Lites , The Staple Singers , Minnie Riperton , Styx ; from the 1980s : Chicago , Cheap Trick , REO Speedwagon , Survivor , Richard Marx ; from the 1990s : R. Kelly ; from the 2000s : Kanye West , Twista , Plain White T 's ."
""",  # noqa: E501
        {
            "Instruction": """I am trying to cluster entity strings on Wikipedia according to the Wikipedia article title they refer to. To help me with this, for a given entity name, please provide me with a comprehensive set of alternative names that could refer to the same entity. Entities may be weirdly truncated or ambiguous - e.g. "Wind" may refer to the band "Earth, Wind, and Fire" or to "rescue service". For each entity, I will provide you with a sentence where this entity is used to help you understand what this entity refers to. Generate a comprehensive set of alternate entity names as a JSON-formatted list.""",  # noqa: E501
            "Demonstrations": """Entity: "fictional character"
Context Sentence: "Jenna Marshall is a fictional character created by Sara Shepard for the `` Pretty Little Liars '' book series , and later developed for the Freeform television series adaptation by I. Marlene King and portrayed by Tammin Sursok ."
Alternate Entity Names: ["fictional characters", "characters", "character"]

Entity: "Catholicism"
Context Sentence: "At home , significantly more electorate residents spoke Italian , Cantonese , Mandarin and Greek at home , and whilst the top three religions (Catholicism , no religion and Anglicanism) differed little from other parts of Perth , Buddhism and Eastern Orthodox adherents outnumbered those of the Uniting Church ."
Alternate Entity Names: ["Catholic Church", "Roman Catholic", "Catholic"]""",  # noqa: E501
        },
    ),
    (
        """You are an expert baker answering users' questions. Reply as agent.

Example conversation:

User: Hey can you help me with something

Agent: Sure! What do you need help with?

User: I want to bake a cake but don't know what temperature to set the oven to.

Agent: For most cakes, the oven should be preheated to 350°F (177°C).

Current conversation:

User: [Insert user's question]

Agent:""",
        {
            "Instruction": (
                "You are an expert baker answering users' "
                + "questions. Reply as agent."
            ),
            "Demonstrations": """User: Hey can you help me with something

Agent: Sure! What do you need help with?

User: I want to bake a cake but don't know what temperature to set the oven to.

Agent: For most cakes, the oven should be preheated to 350°F (177°C).""",
        },
    ),
    (
        "You are given a list of integers. A list is shown by comma-separated numbers between two brackets. For example, [7,3,6] is a list. The number in location one is 7, the number in location two is 3, and the number in location three is 6. You should answer with a list such that every element at each location is equal to the product of elements at every other location in the input array. For example, if a list has four numbers, the answer you give should be created like this: First element of your list = product of second, third, and fourth elements in the given list. Second element of your list = product of First, third and fourth elements in the given list, etc.",  # noqa: E501
        {
            "Instruction": "You are given a list of integers. A list is shown by comma-separated numbers between two brackets. For example, [7,3,6] is a list. The number in location one is 7, the number in location two is 3, and the number in location three is 6. You should answer with a list such that every element at each location is equal to the product of elements at every other location in the input array. For example, if a list has four numbers, the answer you give should be created like this: First element of your list = product of second, third, and fourth elements in the given list. Second element of your list = product of First, third and fourth elements in the given list, etc.",  # noqa: E501
            "Demonstrations": "N/A",
        },
    ),
    (
        "I am learning Japanese. Please translate some Japanese sentences to English. For example, Japanese: その日、人類は思い出した。ヤツらに支配されていた恐怖を鳥籠の中に囚われていた屈辱を English: On that day, humanity remembered the fear of being dominated by them and the humiliation of being trapped in a birdcage.",  # noqa: E501
        {
            "Instruction": "I am learning Japanese. Please translate some Japanese sentences to English.",  # noqa: E501
            "Demonstrations": "Japanese: その日、人類は思い出した。ヤツらに支配されていた恐怖を鳥籠の中に囚われていた屈辱を English: On that day, humanity remembered the fear of being dominated by them and the humiliation of being trapped in a birdcage.",  # noqa: E501
        },
    ),
    (
        "来到美国后，我需要学习如何自己做饭。你能告诉我一些菜需要准备的原料么？这里有一些例子：1. 菜名：西红柿炒蛋。原料：2. 菜名：青椒肉丝炒肉。原料：瘦肉、青椒、调味料（如大蒜、姜、料酒、生抽、盐、糖、鸡精或味精、胡椒粉）、植物油。",  # noqa: E501
        {
            "Instruction": "来到美国后，我需要学习如何自己做饭。你能告诉我一些菜需要准备的原料么？",  # noqa: E501
            "Demonstrations": "2. 菜名：青椒肉丝炒肉。原料：瘦肉、青椒、调味料（如大蒜、姜、料酒、生抽、盐、糖、鸡精或味精、胡椒粉）、植物油。",  # noqa: E501
        },
    ),
    (
        "As a programer, I am learning software development. Here are some of my problems. Input: What is CI/CD? Output: CI/CD is a way to automate and speed up software development by continuously integrating code changes and deploying them quickly and reliably. Input: What is Git? Output:",  # noqa: E501
        {
            "Instruction": "As a programer, I am learning software development. Here are some of my problems.",  # noqa: E501
            "Demonstrations": " Input: What is CI/CD? Output: CI/CD is a way to automate and speed up software development by continuously integrating code changes and deploying them quickly and reliably.",  # noqa: E501
        },
    ),
]


def construct_single_demonstration(
    user_prompt: str,
    parse_dict: dict[str, str] | None,
    input_only: bool = False,
) -> str:
    """Format a demonstration or prediction example to give to an LLM.

    Args:
        user_prompt: A textual prompt describing the task.
        parse_dict: A parsing dictionary containing the correct prompt parse
                    corresponding to the given `user_prompt`.
        input_only: Whether the returned string should only contain
                    the input part. Defaults to False. This should be set to
                    true when constructing the final prediction template to be
                    completed by the LLM.
    """
    input_part = f'''Prompt: """\n{user_prompt}\n"""\n\nParsed Outputs:\n'''
    if input_only:
        return input_part
    output_part = json.dumps(parse_dict, ensure_ascii=False)
    return input_part + output_part


def construct_prompt_for_instruction_parsing(user_prompt: str) -> str:
    """A (GPT-3) prompt for separating instructions from demonstrations.

    Args:
        user_prompt: A user-generated prompt asking for a response.

    Returns:
        A prompt to instruct GPT-3 to parse the user's provided prompt.
    """
    prompt_sections = [METAPROMPT_INSTRUCTION]
    for prompt, correct_parse in METAPROMPT_EXAMPLES:
        prompt_sections.append(
            construct_single_demonstration(prompt, correct_parse, input_only=False)
        )
    all_prompts = "\n\n------\n\n".join(prompt_sections) + "\n\n------\n\n"
    user_input = construct_single_demonstration(user_prompt, None, input_only=True)
    all_prompts += (
        "After seeing these parsed output, please parse this prompt:\n\n" + user_input
    )
    return all_prompts



"""An dual-encoder model retriever using HuggingFace model descriptions."""

from __future__ import annotations  # noqa FI58

import json
import os
import tarfile
import urllib.request

import numpy as np
import retriv
import torch
from tqdm import tqdm

from prompt2model.model_retriever.base import ModelRetriever
from prompt2model.model_retriever.generate_hypothetical_document import (
    generate_hypothetical_model_description,
)
from prompt2model.prompt_parser import PromptSpec
from prompt2model.utils import encode_text, retrieve_objects


class ModelInfo:
    """Store the model name, description, and query-model score for each model."""

    def __init__(
        self,
        name: str,
        description: str,
        score: float,
        size_in_bytes: int,
        num_downloads: int,
    ):
        """Initialize a ModelInfo object.

        Args:
            name: The name of the model.
            description: The description of the model.
            score: The similarity of the model to a given prompt from a user.
            size_in_bytes: The size of the model on disk, in bytes.
            num_downloads: The number of downoads for this model on HuggingFace.
        """
        self.name = name
        self.description = description
        self.score = score
        self.size_in_bytes = size_in_bytes
        self.num_downloads = num_downloads


class DescriptionModelRetriever(ModelRetriever):
    """Retrieve a model from among HuggingFace models."""

    def __init__(
        self,
        search_index_path: str | None = None,
        search_depth: int = 5,
        first_stage_depth: int = 1000,
        encoder_model_name: str = "OpenMatch/cocodr-base-msmarco",
        model_descriptions_index_path="huggingface_models/model_info/",
        device: torch.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        ),
        model_size_limit_bytes=3e9,
        use_bm25: bool = True,
        bm25_index_name: str = "model-index",
        use_HyDE: bool = False,
        openai_api_key: str | None = None,
    ):
        """Initialize a dual-encoder retriever against a search index.

        Args:
            search_index_path: Where to store the search index (e.g. encoded vectors).
            search_depth: The number of most-relevant models to retrieve.
            first_stage_depth: The number of models to retrieve purely by similarity,
                before reranking by scaling with model size and number of downloads
                in the past month.
            encoder_model_name: The name of the model to use for the dual-encoder.
            model_descriptions_index_path: The directory of models to search against.
            device: The device to use for encoding text for our dual-encoder model.
            model_size_limit_bytes: The maximum size (in bytes) of a model to retrieve.
            use_bm25: Whether to use BM25 to retrieve the top-k models. If False, we
                use a dual-encoder retriever.
            bm25_index_name: The name used to save the search index for BM25.
            use_HyDE: Whether to use HyDE to replace the query with a hypothetical
                model description generated by an LLM.
            openai_api_key: OpenAI API key. If None, default to using the
                OPENAI_API_KEY environment variable.
        """
        self.search_depth = search_depth
        self.first_stage_depth = first_stage_depth
        self.encoder_model_name = encoder_model_name
        self.model_descriptions_index_path = model_descriptions_index_path
        self.device = device
        self.model_size_limit_bytes = model_size_limit_bytes
        # If use_bm25 is True, then we use BM25 to retrieve the top-k models.
        # Otherwise, we use a dual-encoder retriever.
        self.use_bm25 = use_bm25
        self.use_HyDE = use_HyDE
        self.openai_api_key = openai_api_key

        # Blocklist certain models' organizations to exclude from model retrieval
        # search results; certain organizations programmatically create models which
        # are unlikely to be useful for task-specific finetuning.
        self.model_blocklist_organizations = ["huggingtweets"]
        self.load_model_info()

        if self.use_bm25:
            assert (
                search_index_path is None
            ), "BM25 expects a search index path with a particular format, so search_index_path should not be provided."  # noqa E501
            self.bm25_index_name = bm25_index_name
            self._search_index_path = retriv.paths.index_path(self.bm25_index_name)
        else:
            if search_index_path is not None:
                assert not os.path.isdir(
                    search_index_path
                ), f"Search index must either be a valid file or not exist yet. But {search_index_path} is provided."  # noqa E501
                self._search_index_path = search_index_path

    @property
    def search_index_path(self):
        """Return the search index path."""
        return self._search_index_path

    def load_model_info(self):
        """Load metadata (e.g. downloads, publication date) about various models.

        We load metadata from json files in the model_descriptions_index_path
        directory, filter out models from certain organizations, and initialize a list
        of ModelInfo objects corresponding to the models we want to search against.
        """
        if not os.path.isdir(self.model_descriptions_index_path):
            # If the model descriptions directory is not populated, then populate it.
            urllib.request.urlretrieve(
                "http://phontron.com/data/prompt2model/model_info.tgz",
                "/tmp/model_info.tgz",
            )
            tar = tarfile.open("/tmp/model_info.tgz")
            os.makedirs(self.model_descriptions_index_path)
            tar.extractall(path=self.model_descriptions_index_path)

        description_files = os.listdir(self.model_descriptions_index_path)
        # We store model names and descriptions in a list of ModelInfo objects.
        self.model_infos: list[ModelInfo] = []
        for f in tqdm(description_files):
            if (
                f.startswith(".")
                or len(open(os.path.join(self.model_descriptions_index_path, f)).read())
                == 0
            ):
                continue
            block = False
            for org in self.model_blocklist_organizations:
                if f.startswith(org + "/"):
                    block = True
                    break
            if block:
                continue
            model_dict = json.load(
                open(os.path.join(self.model_descriptions_index_path, f))
            )
            if model_dict.get("size_bytes", 0) == 0:
                continue
            if "description" not in model_dict:
                continue
            model_name = model_dict["pretrained_model_name"]
            model_info = ModelInfo(
                name=model_name,
                description=model_dict["description"],
                score=None,
                size_in_bytes=model_dict["size_bytes"],
                num_downloads=model_dict.get("downloads", 0),
            )
            self.model_infos.append(model_info)

    def encode_model_descriptions(self, search_index_path) -> np.ndarray:
        """Encode model descriptions into a vector for indexing."""
        model_descriptions = [model.description for model in self.model_infos]
        model_vectors = encode_text(
            self.encoder_model_name,
            text_to_encode=model_descriptions,
            encoding_file=search_index_path,
            device=self.device,
        )
        return model_vectors

    def scale_similarity_score(
        self, model_info: ModelInfo, model_score: float
    ) -> float:
        """Adjust the search score using the model size and number of downloads.

        Args:
            model_info: The name of the model we are scoring.
            model_score: The similarity score of this model for this particular query.
        """
        num_downloads = int(model_info.num_downloads)
        # By taking the log of the number of downloads plus 2, we avoid zeroing the
        # score of models with 0 downloads.
        log_num_downloads = np.log10(num_downloads + 2)
        model_size_bytes = int(model_info.size_in_bytes)
        if model_size_bytes > self.model_size_limit_bytes:
            return -np.inf
        return model_score * log_num_downloads

    def bm25_index_exists(self):
        """Check if a BM25 index exists."""
        assert self.use_bm25 is True
        return (
            os.path.exists(self._search_index_path)
            and len(os.listdir(self._search_index_path)) > 0
        )

    def construct_bm25_index(self, model_infos):
        """Construct a retriv BM25 index for model descriptions."""
        collection = []
        for model in model_infos:
            collection.append({"id": model.name, "text": model.description})
        if self.bm25_index_exists():
            search_engine = retriv.SparseRetriever.load(self._search_index_path)
        else:
            search_engine = retriv.SparseRetriever(self.bm25_index_name)
            search_engine.index(collection)
        return search_engine

    def retrieve(
        self,
        prompt: PromptSpec,
    ) -> list[str]:
        """Select a model from a prompt using a dual-encoder retriever.

        Args:
            prompt: A prompt whose instruction field we use to select relevant models.

        Return:
            A list of relevant models' HuggingFace names.
        """
        if (
            not self.use_bm25
            and self._search_index_path is not None
            and not os.path.exists(self._search_index_path)
        ):
            self.encode_model_descriptions(self._search_index_path)

        if self.use_HyDE:
            query_text = generate_hypothetical_model_description(
                prompt, self.openai_api_key
            )
        else:
            query_text = prompt.instruction

        if self.use_bm25:
            search_engine = self.construct_bm25_index(self.model_infos)
            results = search_engine.search(query_text, cutoff=self.first_stage_depth)
            ranked_list: list[tuple[str, float]] = []
            for result in results:
                ranked_list.append((result["id"], result["score"]))
        else:
            query_vector = encode_text(
                self.encoder_model_name,
                text_to_encode=query_text,
                device=self.device,
            )
            model_names = [model.name for model in self.model_infos]
            ranked_list = retrieve_objects(
                query_vector,
                self._search_index_path,
                model_names,
                self.first_stage_depth,
            )

        model_name_to_model_info = {
            model_info.name: model_info for model_info in self.model_infos
        }

        top_models_list = []
        for model_name, model_score in ranked_list:
            model_info = model_name_to_model_info[model_name]
            scaled_model_score = self.scale_similarity_score(model_info, model_score)
            model_info.score = scaled_model_score
            top_models_list.append(model_info)

        top_models_list = sorted(top_models_list, key=lambda x: x.score, reverse=True)[
            : self.search_depth
        ]
        assert len(top_models_list) > 0, "No models retrieved from search index."
        return [model_info.name for model_info in top_models_list]
